'''
Module to implement robot9000 IRC algorithm.
'''
from hashlib import shake_128 as sha

class BTree:
    def __init__(self):
        self.__nodes = {}

    def insert(self,message,nbytes=8):
        h = int.from_bytes(sha(message).digest(nbytes),'little')
        e = self.__nodes.get(h,0)
        if e:
            return False
        else:
            self.__nodes[h] = 1
            return True

    def __len__(self):
        return len(self.__nodes)

    def __bool__(self):
        return True




'''
    Module to manipulate images, create thumbnails, convert, and add FX.
'''
import subprocess, re, hashlib, os, random

class ImageFileError(Exception):
    pass

# Compilers
ImageFormat = re.compile(r"identify: (no decode delegate for this image format).*")
NotFound = re.compile(r"identify: (unable to open image).*")
Info = re.compile(r"(?P<filename>[/A-Za-z0-9._-]+)(?P<frames>\[\d+\])? (?P<format>[A-Z]+) (?P<size>\d+x\d+).* (?P<filesize>[0-9.]+)([MK]i)?B.*")

def image_info(filename):
    out = subprocess.getoutput('identify '+filename)
    match = Info.match(out)
    if match:
        info = match.groupdict()
        info ['format'] = info['format'].lower()
        info ['size'] = tuple([int(x) for x in info['size'].split('x')])
        if info['frames']:
            info ['frames'] = int(info['frames'].strip('[]'))
        else:
            info.pop ('frames')
        try:
            info ['filesize'] = int(info['filesize'])
        except:
            info ['filesize'] = eval(info['filesize'])
            if match.group(6) == 'Ki':
                info ['filesize'] = int(info['filesize']*(2**10))
            elif match.group(6) == 'Mi':
                info ['filesize'] = int(info['filesize']*(2**20))
        return info
    else:
        for err in [ImageFormat,NotFound]:
            match = err.match(out)
            if match:
                raise ImageFileError(match.group(1))
        raise ImageFileError('Unknow error: `%s`' % out)

def make_thumbnail(original,filename,size):
    return not bool(os.system('convert %s -resize %ix%i %s' % (original,size,size,filename)))

def thumbnail_gif(gif,filename,size):
    return not bool(
            os.system('convert %s -coalesce -resize %ix%i -layers OptimizeFrame %s' % (gif,size,size,filename)))

def resize(original,filename,size):
    return not bool(os.system('convert %s -resize %ix%i! %s' % (original,size[0],size[1],filename)))

def monochrome(original, filename):
    return not bool(os.system('convert %s -colorspace gray -auto-level %s' % (original,filename)))

def checksum(filename,algorithm = 'md5',chunk_size=64*1024):
    Hash = hashlib.new(algorithm)
    with open(filename,'rb') as f:
        chunk = f.read(chunk_size)
        while chunk:
            Hash.update(chunk)
            chunk = f.read(chunk_size)
        f.close()
    return Hash

#Luis Albizo 23/03/18



'''
    This module contains the base classes Post, OP and Reply.
'''
import time

class NoImplementedError(Exception):
    '''This is a way to implement an abstract class.
    '''
    pass

class Post:
    '''The base class for posts.
    '''
    def __init__(self,_id,ip,board,content,image=None,name='Anonymous',option=None):
        self._id = _id
        self._ip = ip
        self._board = board
        self._time = int(time.time())
        self._timestamp = time.localtime()
        self.content = content
        self.option = option
        self.image = image
        self._name = name

    @property
    def id(self):
        return self._id

    @property
    def ip(self):
        return self._ip

    @property
    def time(self):
        return self._time

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def name(self):
        return self._name

    @property
    def board(self):
        return self._board

    def update_time(self):
        self._timestamp = time.localtime()
        self._time = int(time.time())
    
    def move(self):
        raise NoImplementedError('You must to implement your own way to move posts')

    def save(self):
        raise NoImplementedError('You must to implement your own way to save posts')

    def delete(self):
        raise NoImplementedError('You must to implement your own way to delete posts')

class OP(Post):
    '''Original Poster class class..
    '''
    def __init__(self,_id,ip,board,content,image=None,name='Anonymous',title='',option=None):
        super().__init__(_id,ip,board,content,image,name,option)
        self.title=title
        self._replys = []

    @property
    def replys(self):
        return self._replys

class Reply(Post):
    '''Reply class
    '''
    def __init__(self,_id,ip,board,op_id,content,image=None,name='Anonymous',option=None):
        super().__init__(_id,ip,board,content,image,name,option)
        self._op = op_id

    @property
    def op(self):
        return self._op




'''
    ImageBoard Manager for MongoDB implemented with pymongo
'''

from pymongo import MongoClient, MongoReplicaSetClient
from .squema import OP,Reply
from .pqueue import PostQueue as Queue, PostHandler as Handler

def makeID(board):
    '''
        Function to make unique id's in a board.

        :Parameters:
            `board` : a pymongo.collection.Collection instance

        Returns an integer >= 1
    '''
    board.find_one_and_update({'_id':'count'}, {'$inc':{'count':1}}, upsert=True)
    return board.find_one({'_id':'count'}).get('count')

class OP(OP):
    '''
        Original Poster class

        :Parameters:
            `db`        a `pymongo.database.Database` instance
            `_id`       an integer id for the post
            `ip`        the IP addres of the poster
            `board`     the url of the board
            `content`   content of the post 
            `image`     the url of the image (optional, default None)
            `name`      the name of the poster (optional, default is 'Anonymous')
            'title'     a title for the post (optional, default '')
    '''
    def get(db,board,_id):
        '''
        'Static' method of OP
        Not need to instance the class to use.

        :Parameters:
            `db`    a pymongo.database.Database instance
            `board` the url of the board
            `_id`   the unique id of the post
        
        Returns a OP instance if exists
        '''
        op = OP(db,_id,'',board,'')
        if op.load():
            return op

    def __init__(self,db,*args,**kwargs):
        super().__init__(*args,**kwargs)
        self._db = db

    @property
    def replys(self):
        thread = self._db.get_collection(self.board+'.'+str(self.id))
        for r in thread.find({'reply':True},{'_id':True}):
            yield self.get_reply(r['_id'])

    @property
    def replys_id(self):
        return [r['_id'] for r in
                self._db.get_collection(self.board+'.'+str(self.id)).find(
                    {'reply':True},{'_id':True})]
    
    @property
    def info(self):
        info = {
            'OP' : True,
            '_id' : self.id,
            'ip' : self.ip,
            'time' : self.time,
            'timestamp' : self.timestamp,
            'board' : self.board,
            'content' : self.content,
            'name' : self.name
        }
        if self.title:
            info['title'] = self.title
        if self.image:
            info['image'] = self.image
        return info

    def get_reply(self,_id):
        '''
        :Parameters:
            `_id` a unique _id of the reply

        Return a Reply instance if the reply exists
        '''
        match = self._db.get_collection(self.board+'.'+str(self.id)).find_one({'_id':_id},{'reply':False,'time':False,'timestamp':False})
        if match:
            match ['op_id'] = match ['op']
            match.pop('op')
            return Reply(self._db,**match)

    def save(self):
        '''
        Insert or update if exists in database
        '''
        thread = self._db.get_collection(self.board+'.'+str(self.id))
        thread.find_one_and_update({'_id':self.id}, {'$set':self.info}, upsert=True)
    
    def move(self,board,new_id,map_id):
        '''
        Move OP post and his replys to another board.
        :Parameters:
            `board`     the url of board to move the post, must not be the actual board
            `new_id`    a new _id to identify the post in the new board
            `map_id`    a function to map all the reply's id's

        Return the new id of the post
        '''
        assert self.board != board, "can't move a thread to his previous board"
        replys = zip(list(self.replys),map(map_id,self.replys_id))
        self.delete()
        info = {
            'moved' : True,
            '_id' : self.id,
            'new_board' : board,
            'new_id' : new_id
        }
        self._db.get_collection(self.board).insert_one(info)
        self._board = board
        self._id = new_id
        for reply in replys:
            reply[0].move(board=self.board,op_id=self.id,new_id=reply[1])
        self.save()
        return self.id

    def delete(self):
        '''
        Delete the post with all his replys
        '''
        self._db.get_collection(self.board+'.'+str(self.id)).drop()

    def load(self):
        '''
        Update the instance attributes according to the database
        '''
        match = self._db.get_collection(self.board+'.'+str(self._id)).find_one({'OP':True})
        if match:
            self._id = match['_id']
            self._ip = match['ip']
            self._board = match['board']
            self._time = match ['time']
            self.title = match.get('title',None)
            self._timestamp = match['timestamp']
            self.content = match.get('content',None)
            self.image  = match.get('image',None)
            self._name = match.get('name','Anonymous')
            self.option = match.get('option',None)
            return True

class Reply(Reply):
    '''
        Reply class

        :Parameters:
            `db`        a `pymongo.database.Database` instance
            `_id`       an integer id for the post
            `ip`        the IP addres of the poster
            `board`     the url of the board
            `op_id`     the OP._id of the OP for this reply
            `content`   content of the post 
            `image`     the url of the image (optional, default None)
            `name`      the name of the poster (optional, default is 'Anonymous')
            `option`    an extra option option (optional, default None)
    '''
    def __init__(self,db,*args,**kwargs):
        super().__init__(*args,**kwargs)
        self._db=db

    def get(db,board,_id):
        '''
        'Static' method of Reply
        Not need to instance the class to use.

        :Parameters:
            `db`    a pymongo.database.Database instance
            `board` the url of the board
            `_id`   the unique id of the post
        
        Returns a Reply instance if exists
        '''
        for thread in filter(lambda x: x.split('.',1)[0]==board,db.collection_names()):
            match = db.get_collection(thread).find_one({'_id':_id,'reply':True},{'_id':True,'board':True})
            if match:
                with Reply(db,match['_id'],'',match['board'],'','') as reply:
                    reply.load()
                    return reply

    @property
    def info(self):
        info = {
            'reply'     :   True,
            '_id'       :   self.id,
            'ip'        :   self.ip,
            'time'      :   self.time,
            'timestamp' :   self.timestamp,
            'board'     :   self.board,
            'op'        :   self.op,
            'content'   :   self.content,
            'name'      :   self.name,
            'option'    :   self.option
        }
        if self.image:
            info['image'] = self.image
        return info

    def save(self):
        '''
        Insert or update if exists in database
        '''
        thread = self._db.get_collection(self.board+'.'+str(self.op))
        thread.find_one_and_update({'_id':self.id}, {'$set':self.info}, upsert=True)

    def move(self,board,op_id,new_id):
        '''
        Move Reply post to another thread.
        :Parameters:
            `board`     the url of board to move the post, must not be the actual board
            `op_id`     the id of the OP, must exists
            `new_id`    a new _id to identify the post in the new thread

        Return the new id of the post
        '''
        assert self.board != board or self.op != op_id, "can't move a reply to his previous thread"
        self.delete()
        self._op = op_id
        self._board = board
        self._id = new_id
        self.update_time()
        self.save()

    def delete(self):
        '''
        Deletes the entire thread
        '''
        thread = self._db.get_collection(self.board+'.'+str(self.op))
        thread.delete_one({'_id':self.id})
        #thread.drop()
        info = {
            'deleted' : True,
            '_id' : self.id
        }
        self._db.get_collection(self.board+'.'+str(self.op)).insert_one(info)

    def load(self):
        '''
        Update the instance attributes according to the database
        '''
        match = None
        if self.op:
            match = self._db.get_collection(self.board+'.'+str(self.op)).find_one({'_id':self.id,'reply':True})
        else:
            for thread in filter(lambda x: x.split('.')[0]==self.board,self._db.collection_names()):
                match = self._db.get_collection(thread).find_one({'_id':self.id,'reply':True})
                if match:
                    break
        if match:
            self._id = match['_id']
            self._ip = match['ip']
            self._board =  match['board']
            self._time = match['time']
            self._timestamp = match['timestamp']
            self._op = match['op']
            self.content = match.get('content',str())
            self.image = match.get('image',None)
            self._name = match.get('name','Anonymous')
            self.option = match.get('option',None)
            return True

class Manager:
    '''
    A class to manage the database/squema part of an imageboard and the PostQueue.

    :Parameters: (optional)
        `dbname`    the name of the database (default 'imageboard')
        `host`      the hostname of the mongo database (default '127.0.0.1')
        `port`      the port of the mongo database (default 27017)
    '''
    def __init__(self,dbname='imageboard',host='127.0.0.1',port=27017):
        self.__db = MongoClient(host=host, port=port).get_database(dbname)
        self.__handlers = dict()
        self.__queues = dict()

    @property
    def db(self):
        return self.__db

    def handler(self,afunc):
        '''
        Decorator for post handlers.

        Pass a handler function to handle posts.

        Example:
        >>> @manager.handler
        ... def simple_handler(post):
        ...     #do something with post
        ... 
        '''
        self.__handlers[afunc.__name__] = Handler()(afunc)
        return afunc.__name__

    def add_queue(self,name,ratio,handler_name):
        '''
        Add a PostQueue.

        :Parameters:
            `name`      the name of the queue
            `ratio`     the ratio limit of the queue
            `handler_name`      the name of the handler function
        '''
        self.__queues[name] = Queue(ratio,self.__handlers[handler_name])

    def start_queue(self,name):
        '''
        Start a queue by name.
        '''
        self.__queues[name].start()

    def stop_queue(self,name):
        '''
        Stop a queue by name.
        '''
        self.__queues[name].stop()

    def start_all(self):
        '''
        Start all queue's.

        Raises HandleError
        '''
        for q in self.__queues.values():
            q.start()

    def stop_all(self):
        '''
        Stop all queue's.
        '''
        for q in self.__queues.values():
            q.stop()

    def put(self,post,qname):
        '''
        Put post in queue by qname.

        Raises EnqueueError.
        '''
        self.__queues[qname].put(post)

    def create_board(self,url,name,description,**kwargs):
        '''
        Create a board.

        :Parameters:
            `url`   the shortname of the board (this is gonna be the name of the collection)
            `name`  the fullname of the board
            `description`   a short description for the board
            **kwargs for more info if it's necessary
        
        Returns True if the boards not exist before, else returns False.
        '''
        if not(self.exist_board(url)):
            board = {
                'board' : True,
                'url' : url,
                'name' : name,
                'description' : description
            }
            [board.update([(k,kwargs[k])]) for k in kwargs]
            self.__db.get_collection(url).insert_one(board)
            self.__db.boards.insert_one(board)
            return True
        else:
            return False

    def get_board(self,url):
        return self.__db.get_collection(url).find_one({'board':True})

    def delete_board(self,url):
        '''
        Deletes a board.

        :Parameters:
            `url`   the shortname of the board (the name of the collection)
        '''
        self.__db.get_collection(url).drop()

    def exist_board(self,url):
        '''
        Checks if a board exists.

        :Parameters:
            `url`   the shortname of the board (the name of the collection)

        Return True if exists, else False
        '''
        return (url in [c['name'] for c in self.__db.list_collections()])

    def exist_thread(self,board,_id):
        '''
        Checks if a thread exists.

        :Parameters:
            `board`     the shortname of the board (the name of the collection)
            `_id`       the id of the OP

        Return True if exists, else False
        '''
        if not(OP.get(self.__db,board,_id)): 
            return False
        else:
            return True

    def get_thread(self,board,_id):
        '''
        Get a thread basic information.

        :Parameters:
            `board`     the shortname of the board (the name of the collection)
            `_id`       the id of the OP
        '''
        op = OP.get(self.__db,board,_id)
        if not(op): 
            return
        return {
            'op' : op.info,
            'replys_id' : op.replys_id,
            'replys' : op.replys
        }






'''
Module for handle posts.
'''
import threading

class EnqueueError(Exception):
    pass

class HandleError(Exception):
    pass

class PostQueue:
    '''
    A queue to handle posts.
    '''
    def __init__(self,ratiolimit,handle):
        self.__posts = []
        self.ratio = int(ratiolimit)
        assert '__call__' in dir(handle), 'handle must be callable'
        self.__handle = handle
        self.running = False

    @property
    def size(self):
        return len(self.__posts)

    def put(self,post):
        '''
        Add a post at the end of the queue
 
        Raises EnqueueError
        '''
        if self.size >= self.ratio:
            raise EnqueueError('ratio limit reached')
        else:
            self.__posts.append(post)

    def get(self):
        '''
        Get the first element of the queue
        '''
        return self.__posts.pop(0)

    def start(self):
        '''
        Start in a new thread the post handler
        
        Raises HandleError
        '''
        if self.running:
            raise HandleError("handler is already running")
        self.running = True
        threading.Thread(target=self.__handle, args=(self,), daemon=True).start()

    def stop(self):
        '''
        Stop executing the handler
        '''
        self.running = False

class PostHandler:
    '''
    Decorator to make post handler's.

    Usage:
        >>> @PostHandler()
        ... def handler(post):
        ...     #do something
        ...     pass
        ... 
        >>> queue = PostQueue(10,handler)
    '''
    def __init__(self):
        pass

    def __call__(self,handler):
        def new_handler(queue):
            while queue.running:
                if queue.size:
                    handler(queue.get())
        return new_handler




__author__ = 'Luis Albizo'
__license__ = 'Creative Commons Attribution-ShareAlike 4.0 International License'
__email__ = 'albizo.luis@gmail.com'
__date__ = '26 feb 2018'
__version__ = '0.3.2'

import milachan.squema
import milachan.pqueue

class managers:
    @property
    def MongoManager(_):
        __import__('milachan.manager-pymongo')
        return getattr(milachan,'manager-pymongo')

managers = managers()




from setuptools import setup
from os import path

setup(
  name = 'milachan',
  packages = ['milachan'],
  version = '0.3.2',
  description = 'A minimalistic framework for creating simple ImageBoards',
  author = 'Luis Albizo',
  author_email = 'albizo.luis@gmail.com',
  url = 'https://github.com/LuisAlbizo/milachan',
  keywords = ['imageboard', 'framework', 'chan','backend','database'],
  license = 'Creative Commons Attribution-ShareAlike 4.0 International License',
  setup_cfg = True,
  classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: Other/Proprietary License ',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ]
)

